import boto3
import pytest

from nops_metadata import METAMAP
from nops_metadata import MetaFetcher


def test_meta_fetcher_init(metafetcher: MetaFetcher):
    assert metafetcher


def test_metamap_naming():
    """Metamap naming must be strongly consistent"""
    for key, element in METAMAP.items():
        assert element is not None
        _, entity_name = key.split("_", 1)

        fetch_method = element["fetch_method"]
        assert isinstance(fetch_method, str)
        _, entity_from_fetch_method = fetch_method.split("_", 1)
        assert entity_name == entity_from_fetch_method


def test_meta_fetcher_region_list(metafetcher: MetaFetcher):
    assert metafetcher.metadata_regions(metadata_type="ec2_instances")


def test_meta_fetcher_fetch_method(metafetcher: MetaFetcher):
    assert metafetcher.fetch_method_name(metadata_type="ec2_instances") == "describe_instances"

    for metadata_type in metafetcher.metadata_types:
        assert all(metafetcher.fetch_method_name(metadata_type=metadata_type))

    for subresource_metadata_type in metafetcher.subresources_metadata_types:
        assert all(metafetcher.fetch_method_name(subresource_metadata_type=subresource_metadata_type))


def test_meta_fetcher_service_name(metafetcher: MetaFetcher):
    assert metafetcher.fetch_method_service(metadata_type="ec2_instances") == "ec2"

    for metadata_type in metafetcher.metadata_types:
        assert metafetcher.fetch_method_service(metadata_type=metadata_type) == metadata_type.split("_")[0]

    for subresource_metadata_type in metafetcher.subresources_metadata_types:
        assert (
            metafetcher.fetch_method_service(subresource_metadata_type=subresource_metadata_type)
            == subresource_metadata_type.split("_")[0]
        )


def test_meta_fetcher_region_list_wrong_name(metafetcher: MetaFetcher):
    assert not metafetcher.metadata_regions(metadata_type="unknown_servicecall")


@pytest.fixture
def ecs_clusters(metafetcher: MetaFetcher):
    resources = metafetcher.fetch(metadata_type="ecs_clusters", region_name="us-west-2")
    return [r for r in resources]


@pytest.fixture
def ec2_instances(metafetcher: MetaFetcher):
    resources = metafetcher.fetch(metadata_type="ec2_instances", region_name="us-west-2")
    return [r for r in resources]


@pytest.fixture
def elbv2_target_groups(metafetcher: MetaFetcher):
    resources = metafetcher.fetch(metadata_type="elbv2_target_groups", region_name="us-west-2")
    return [r for r in resources]


def test_meta_fetcher_pulling_listing(ecs_clusters):
    assert ecs_clusters
    assert isinstance(ecs_clusters, list)

    for element in ecs_clusters:
        assert isinstance(element, dict)

    assert all(ecs_clusters)


def test_meta_fetcher_metadata_types(metafetcher):
    """Confirm that all calls are working"""
    for metadata_type in metafetcher.metadata_types:
        assert all(metafetcher.fetch(metadata_type=metadata_type, region_name="eu-central-1"))


def test_meta_fetcher_schema_export(metafetcher: MetaFetcher):
    for metadata_type in metafetcher.metadata_types:
        spark_schema = metafetcher.schema(metadata_type=metadata_type)
        postgres_schema = metafetcher.postgres_schema(metadata_type=metadata_type)
        assert spark_schema
        assert postgres_schema


def test_meta_fetcher_pulling_resource_details_payload(metafetcher: MetaFetcher, elbv2_target_groups: list[dict]):
    for resource_details_type in metafetcher.subresources_metadata_types:
        if resource_details_type != "elbv2_target_health":
            continue

        parent_filters = []
        metadata_config = metafetcher.metadata_config(metadata_type="elbv2_target_health")
        assert metadata_config
        filter_key = metadata_config["parent_required_filters"]["filter_key"]
        parent_field_name = metadata_config["parent_required_filters"]["parent_filter_field"]

        for resource in elbv2_target_groups[:2]:
            parent_filters.append({filter_key: resource[parent_field_name]})

        assert len(parent_filters) == 2

        resources = metafetcher.fetch(
            metadata_type="ecs_clusters", region_name="us-west-2", required_filters=parent_filters
        )
        for detail in resources:
            assert detail
            assert isinstance(detail, dict)


def test_meta_fetcher_custom_kwargs(metafetcher: MetaFetcher, ec2_instances):

    assert ec2_instances
    assert isinstance(ec2_instances, list)

    resources_list = []

    ec2_instances = ec2_instances[:2]

    for instance in ec2_instances:
        assert isinstance(instance, dict)
        instance_id = instance["InstanceId"]
        custom_kwargs = {"InstanceIds": [instance_id]}
        resources = metafetcher.fetch(
            metadata_type="ec2_instances", region_name="us-west-2", custom_kwargs=custom_kwargs
        )
        assert resources

        for r in resources:
            assert isinstance(r, dict)
            resources_list.append(r["InstanceId"])
            assert r["InstanceId"] == instance_id

    assert len(resources_list) == len(ec2_instances)



import logging
from threading import Thread
from typing import Any
from typing import Iterator
from typing import Optional
from datetime import datetime

import boto3
from pyrsistent import thaw

from .constants import METAMAP
from .constants import SUBRESOURCES_METAMAP
from .constants import SINLE_REGION_SERVICES
from .schema import PydanticSchemaGenerator
from .spark_schema import SparkAWSSchemaSerializer
from .utils import resource_listing
from queue import Empty
from typing import List
from queue import Queue


class MetaFetcher:
    def __init__(self, session: boto3.Session):
        self.session: boto3.Session = session

    def metadata_regions(self, metadata_type: str) -> list[str]:
        service_name = metadata_type.split("_")[0]
        if service_name in SINLE_REGION_SERVICES:
            return ["us-east-1"]
        else:
            return self.session.get_available_regions(service_name=service_name)

    def fetch_method_name(self, metadata_type: str = "", subresource_metadata_type: str = "") -> str:
        if metadata_type:
            return METAMAP[metadata_type]["fetch_method"]
        elif subresource_metadata_type:
            return SUBRESOURCES_METAMAP[subresource_metadata_type]["fetch_method"]
        else:
            raise ValueError

    def fetch_method_service(self, metadata_type: str = "", subresource_metadata_type: str = "") -> str:
        if metadata_type:
            return metadata_type.split("_")[0]
        elif subresource_metadata_type:
            return subresource_metadata_type.split("_")[0]
        else:
            raise ValueError

    def schema(self, metadata_type: str, schema_type: str = "spark") -> Any:
        if schema_type == "spark":
            return SparkAWSSchemaSerializer().schema(metadata_type=metadata_type)
        elif schema_type == "pydantic":
            return PydanticSchemaGenerator().schema(metadata_type)
        else:
            raise ValueError(f"{schema_type} is not supported.")

    def postgres_schema(self, metadata_type: str) -> dict[str, Any]:
        postgres_custom_types = {
            "string": "varchar",
            "long": "integer",
            "double": "float",
        }
        serializer = SparkAWSSchemaSerializer(custom_types=postgres_custom_types)
        schema = serializer.schema(metadata_type)
        return schema

    def metadata_config(self, metadata_type: str) -> dict[str, Any]:
        return (METAMAP | SUBRESOURCES_METAMAP)[metadata_type]

    @property
    def metadata_types(self) -> list[str]:
        return list(METAMAP.keys())

    @property
    def subresources_metadata_types(self) -> list[str]:
        return list(SUBRESOURCES_METAMAP.keys())

    def fetch_in_threads(
        self,
        metadata_config: dict,
        call_kwargs: dict,
        metadata_type: str,
        region_name: str,
        num_threads: int,
        required_filters: Optional[list[dict[str, Any]]] = None,
    ):
        kwargs_list = []
        for filter_kwargs in required_filters or []:
            kwargs_list.append(dict(call_kwargs, **filter_kwargs))

        def _worker():
            while True:
                try:
                    task_kwargs = task_queue.get(block=False)
                    resources = resource_listing(
                        session=self.session,
                        metaname=metadata_type,
                        fetch_method=metadata_config["fetch_method"],
                        response_key=metadata_config.get("response_key"),
                        page_key=metadata_config.get("page_key", ""),
                        call_kwargs=task_kwargs,
                        region_name=region_name,
                    )
                    queue.put(list([dict(resource, **task_kwargs) for resource in resources]), timeout=60 * 3)

                # Ends worker life.
                except Empty:
                    break

                except Exception as e:
                    pass

        task_queue: Queue = Queue(maxsize=len(kwargs_list))
        queue: Queue = Queue(maxsize=len(kwargs_list))

        for kwarg in kwargs_list:
            task_queue.put(kwarg)

        threads: List[Thread] = [Thread(target=_worker) for _ in range(num_threads)]

        for thread in threads:
            thread.start()

        while True:
            try:
                yield from queue.get(block=True, timeout=5)

            except Empty:
                any_alive = any([thread.is_alive() for thread in threads])
                if not any_alive:
                    break

    def fetch(
        self,
        metadata_type: str,
        region_name: str,
        required_filters: Optional[list[dict[str, Any]]] = None,
        custom_kwargs: Optional[dict[str, Any]] = None,
        num_threads: int = 5,
    ) -> Iterator[dict[str, Any]]:
        metadata_config = self.metadata_config(metadata_type)
        call_kwargs = thaw(metadata_config.get("kwargs", {}))

        if "parent_required_filters" in metadata_config:
            yield from self.fetch_in_threads(
                metadata_config=metadata_config,
                call_kwargs=call_kwargs,
                metadata_type=metadata_type,
                region_name=region_name,
                required_filters=required_filters,
                num_threads=num_threads,
            )
        else:
            yield from resource_listing(
                session=self.session,
                metaname=metadata_type,
                fetch_method=metadata_config["fetch_method"],
                response_key=metadata_config.get("response_key"),
                page_key=metadata_config.get("page_key", ""),
                call_kwargs=custom_kwargs or call_kwargs,
                region_name=region_name,
            )



from collections import OrderedDict

from pyrsistent import freeze

SINLE_REGION_SERVICES = freeze(["iam", "s3", "cloudfront"])

# fmt: off
METAMAP = freeze(
    OrderedDict({
        "autoscaling_auto_scaling_groups": {"fetch_method": "describe_auto_scaling_groups", "response_key": "AutoScalingGroups"},
        "cloudformation_stacks": {"fetch_method": "describe_stacks", "response_key": "Stacks"},
        "cloudfront_distributions": {"fetch_method": "list_distributions", "response_key": "Items", "page_key": "DistributionList"},
        "cloudtrail_trails": {"fetch_method": "describe_trails", "response_key": "trailList"},
        "config_configuration_recorder_status": {"fetch_method": "describe_configuration_recorder_status", "response_key": "ConfigurationRecordersStatus"},
        "config_configuration_recorders": {"fetch_method": "describe_configuration_recorders", "response_key": "ConfigurationRecorders"},
        "config_delivery_channel_status": {"fetch_method": "describe_delivery_channel_status", "response_key": "DeliveryChannelsStatus"},
        "config_delivery_channels": {"fetch_method": "describe_delivery_channels", "response_key": "DeliveryChannels"},
        "dynamodb_tables": {"fetch_method": "list_tables", "response_key": "TableNames"},
        "ec2_addresses": {"fetch_method": "describe_addresses", "response_key": "Addresses"},
        "ec2_flow_logs": {"fetch_method": "describe_flow_logs", "response_key": "FlowLogs"},
        "ec2_images": {"fetch_method": "describe_images", "response_key": "Images", "kwargs": {"Owners": ["self"]}},
        "ec2_instances": {"fetch_method": "describe_instances", "page_key": "Reservations", "response_key": "Instances"},
        "ec2_nat_gateways": {"fetch_method": "describe_nat_gateways", "response_key": "NatGateways"},
        "ec2_network_interfaces": {"fetch_method": "describe_network_interfaces", "response_key": "NetworkInterfaces"},
        "ec2_reserved_instances": {"fetch_method": "describe_reserved_instances", "response_key": "ReservedInstances"},
        "ec2_reserved_instances_modifications": {"fetch_method": "describe_reserved_instances_modifications", "response_key": "ReservedInstancesModifications"},
        "ec2_route_tables": {"fetch_method": "describe_route_tables", "response_key": "RouteTables"},
        "ec2_security_groups": {"fetch_method": "describe_security_groups", "response_key": "SecurityGroups"},
        "ec2_snapshots": {"fetch_method": "describe_snapshots", "response_key": "Snapshots", "kwargs": {"OwnerIds": ["self"]}},
        "ec2_subnets": {"fetch_method": "describe_subnets", "response_key": "Subnets"},
        "ec2_volumes": {"fetch_method": "describe_volumes", "response_key": "Volumes"},
        "ec2_vpcs": {"fetch_method": "describe_vpcs", "response_key": "Vpcs"},
        "ecs_clusters": {"fetch_method": "list_clusters", "response_key": "clusterArns"},
        "efs_file_systems": {"fetch_method": "describe_file_systems", "response_key": "FileSystems"},
        "eks_clusters": {"fetch_method": "list_clusters", "response_key": "clusters"},
        "elasticache_cache_clusters": {"fetch_method": "describe_cache_clusters", "response_key": "CacheClusters"},
        "elasticache_replication_groups": {"fetch_method": "describe_replication_groups", "response_key": "ReplicationGroups"},
        "elasticache_cache_subnet_groups": {"fetch_method": "describe_cache_subnet_groups", "response_key": "CacheSubnetGroups"},
        "elbv2_load_balancers": {"fetch_method": "describe_load_balancers", "response_key": "LoadBalancers"},
        "elbv2_target_groups": {"fetch_method": "describe_target_groups", "response_key": "TargetGroups"},
        "iam_account_aliases": {"fetch_method": "list_account_aliases", "response_key": "AccountAliases"},
        "iam_account_summary": {"fetch_method": "get_account_summary", "response_key": "SummaryMap"},
        "iam_account_password_policy": {"fetch_method": "get_account_password_policy", "response_key": "PasswordPolicy"},
        "iam_instance_profiles": {"fetch_method": "list_instance_profiles", "response_key": "InstanceProfiles"},
        "iam_roles": {"fetch_method": "list_roles", "response_key": "Roles"},
        "iam_users": {"fetch_method": "list_users", "response_key": "Users"},
        "inspector_assessment_runs": {"fetch_method": "list_assessment_runs", "response_key": "assessmentRunArns"},
        "kms_keys": {"fetch_method": "list_keys", "response_key": "Keys"},
        "lambda_functions": {"fetch_method": "list_functions", "response_key": "Functions"},
        "neptune_db_instances": {"fetch_method": "describe_db_instances", "response_key": "DBInstances"},
        "rds_db_instances": {"fetch_method": "describe_db_instances", "response_key": "DBInstances"},
        "rds_db_snapshots": {"fetch_method": "describe_db_snapshots", "response_key": "DBSnapshots"},
        "rds_pending_maintenance_actions": {"fetch_method": "describe_pending_maintenance_actions", "response_key": "PendingMaintenanceActions"},
        "redshift_clusters": {"fetch_method": "describe_clusters", "response_key": "Clusters"},
        "resourcegroupstaggingapi_tag_keys": {"fetch_method": "get_tag_keys", "response_key": "TagKeys"},
        "resourcegroupstaggingapi_resources": {"fetch_method": "get_resources", "response_key": "ResourceTagMappingList"},
        "s3_buckets": {"fetch_method": "list_buckets", "response_key": "Buckets"},
        "ssm_compliance_summaries": {"fetch_method": "list_compliance_summaries", "response_key": "ComplianceSummaryItems"},
        "workspaces_workspace_directories": {"fetch_method": "describe_workspace_directories", "response_key": "Directories"},
        "workspaces_workspaces": {"fetch_method": "describe_workspaces", "response_key": "Workspaces"},
    })
)

SUBRESOURCES_METAMAP = freeze(
    OrderedDict({
        "elbv2_target_health": {
            "fetch_method": "describe_target_health",
            "response_key": "TargetHealthDescriptions",
            "parent_required_filters": {
                "filter_key": "TargetGroupArn",
                "parent_filter_field": "TargetGroupArn",
                "parent_metadata_type": "elbv2_target_groups",
            }
        },
        "s3_bucket_versioning": {
            "fetch_method": "get_bucket_versioning",
            "parent_required_filters": {
                "filter_key": "Bucket",
                "parent_filter_field": "Name",
                "parent_metadata_type": "s3_buckets",
            }
        },
        "s3_bucket_encryption": {
            "fetch_method": "get_bucket_encryption",
            "response_key": "ServerSideEncryptionConfiguration",
            "parent_required_filters": {
                "filter_key": "Bucket",
                "parent_filter_field": "Name",
                "parent_metadata_type": "s3_buckets",
            }
        }
    })
)

# This dict stores the "metadata_type -> fields to process" mapping
# We need it because sweet boto3 lib marks AssumeRolePolicyDocument as str in schemas, but in fact it returns dict obj.
# SEE: nops_metadata.utils.process_element
APPLY_NESTED_ELEMENT_PROCESSING = {
    "iam_instance_profiles": ["Roles"]
}

RELATIONSHIPS_MAPPING = freeze(
    {
        "ec2_instances": {
            "volume": {
                "many_to_many": True,
                "m2m_table_name": "ec2_instance_volume",
                "related_table": "ec2_volumes",
                "related_column": "volume_id",
                "custom_join_query": """, json_array_elements(block_device_mappings::json) AS e JOIN ec2_volumes ec2_volumes_tmp ON (e -> 'Ebs' ->> 'VolumeId') = ec2_volumes_tmp.volume_id """,
            },
            "image_fk": {
                "related_table": "ec2_images",
                "related_column": "image_id",
            },
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            },
            "subnet_fk": {
                "related_table": "ec2_subnets",
                "related_column": "subnet_id",
            },
            "ec2_security_groups": {
                "many_to_many": True,
                "m2m_table_name": "ec2_instance_security_group",
                "related_table": "ec2_security_groups",
                "related_column": "group_id",
                "custom_join_query": """, json_array_elements(security_groups::json) AS e JOIN ec2_security_groups ec2_security_groups_tmp ON (e ->> 'GroupId') = ec2_security_groups_tmp.group_id """,
             },
             "iam_instance_profile_fk": {
                "related_table": "iam_instance_profiles",
                "related_column": "instance_profile_id",
                "custom_join_query": """LEFT OUTER JOIN iam_instance_profiles iam_instance_profiles_tmp ON iam_instance_profiles_tmp.arn = vals.iam_instance_profile->>'Arn'""",
             },
         },
         "iam_instance_profiles": {
            "roles": {
                "many_to_many": True,
                "m2m_table_name": "iam_instance_profile_role",
                "related_table": "iam_roles",
                "related_column": "arn",
                "custom_join_query": """, json_array_elements(roles::json) AS e JOIN iam_roles iam_roles_tmp ON (e ->> 'Arn') = iam_roles_tmp.arn """,
            }
        },
        "ec2_security_groups": {
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            },
        },
        "elbv2_target_groups": {
            "load_balancer_arns": {
                "many_to_many": True,
                "m2m_table_name": "target_group_load_balancer",
                "related_table": "elbv2_load_balancers",
                "related_column": "load_balancer_arn",
            },
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            },
        },
        "elbv2_load_balancers": {
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            },
            "security_groups": {
                "many_to_many": True,
                "m2m_table_name": "load_balancer_security_group",
                "related_table": "ec2_security_groups",
                "related_column": "group_id",
            }
        },
        "elbv2_target_health": {
            "instance_fk": {
                "related_table": "ec2_instances",
                "custom_join_query": """LEFT OUTER JOIN ec2_instances ec2_instances_tmp ON ec2_instances_tmp.instance_id = vals.target->>'Id'""",
                "related_column": "instance_id",
            }
        },
        "ec2_snapshots": {
            "volume_fk": {
                "related_table": "ec2_volumes",
                "related_column": "volume_id",
            }
        },
        "ec2_addresses": {
            "instance_fk": {
                "related_table": "ec2_instances",
                "related_column": "instance_id",
            },
            "network_interface_fk": {
                "related_table": "ec2_network_interfaces",
                "related_column": "network_interface_id",
            }
        },
        "ec2_route_tables": {
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            }
        },
        "ec2_subnets": {
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            }
        },
        "rds_db_snapshots": {
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            }
        },
        "ec2_flow_logs": {
            "resource_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
                "this_table_column": "resource_id"
            }
        },
        "ec2_nat_gateways": {
            "subnet_fk": {
                "related_table": "ec2_subnets",
                "related_column": "subnet_id",
            },
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            }
        },
        "autoscaling_auto_scaling_groups": {
            "instances": {
                "many_to_many": True,
                "m2m_table_name": "autoscaling_enabled_ec2_instance",
                "related_table": "ec2_instances",
                "related_column": "instance_id",
                "custom_join_query": """, json_array_elements(instances::json) AS e JOIN ec2_instances ec2_instances_tmp ON (e ->> 'InstanceId') = ec2_instances_tmp.instance_id """,
            }
        },
        "rds_db_instances": {
            "kms_key_pk": {
                "related_table": "kms_keys",
                "related_column": "key_arn",
                "this_table_column": "kms_key_id",
            },
            "vpc_security_groups": {
                "many_to_many": True,
                "m2m_table_name": "rds_instance_security_group",
                "related_table": "ec2_security_groups",
                "related_column": "group_id",
                "custom_join_query": """, json_array_elements(vpc_security_groups::json) AS e JOIN ec2_security_groups ec2_security_groups_tmp ON (e ->> 'VpcSecurityGroupId') = ec2_security_groups_tmp.group_id """,
            }
        },
        "efs_file_systems": {
            "kms_key_pk": {
                "related_table": "kms_keys",
                "related_column": "key_arn",
                "this_table_column": "kms_key_id",
            },
        },
        "elasticache_cache_clusters": {
            "security_groups": {
                "many_to_many": True,
                "m2m_table_name": "elasticache_security_group",
                "related_table": "ec2_security_groups",
                "related_column": "group_id",
                "custom_join_query": """, json_array_elements(security_groups::json) AS e JOIN ec2_security_groups ec2_security_groups_tmp ON (e ->> 'SecurityGroupId') = ec2_security_groups_tmp.group_id """,
            },
            "replication_group_pk": {
                "related_table": "elasticache_replication_groups",
                "related_column": "replication_group_id",
            },
            "cache_subnet_group_name_fk": {
                "related_table": "elasticache_cache_subnet_groups",
                "related_column": "cache_subnet_group_name",
            }
        },
        "elasticache_replication_groups": {
            "kms_key_pk": {
                "related_table": "kms_keys",
                "related_column": "key_arn",
                "this_table_column": "kms_key_id",
            },
        },
        "elasticache_cache_subnet_groups": {
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            },
            "subnets": {
                "many_to_many": True,
                "m2m_table_name": "elasticache_cache_subnet_group",
                "related_table": "ec2_subnets",
                "related_column": "subnet_id",
                "custom_join_query": """, json_array_elements(subnets::json) AS e LEFT OUTER JOIN ec2_subnets ec2_subnets_tmp ON (e ->> 'SubnetIdentifier') = ec2_subnets_tmp.subnet_id """,
            },
        },
        "redshift_clusters": {
            "vpc_fk": {
                "related_table": "ec2_vpcs",
                "related_column": "vpc_id",
            },
            "kms_key_pk": {
                "related_table": "kms_keys",
                "related_column": "key_arn",
                "this_table_column": "kms_key_id",
            },
        },
        "cloudtrail_trails": {
            "s3_bucket_fk": {
                "related_table": "s3_buckets",
                "related_column": "name",
                "this_table_column": "s3_bucket_name",
            },
            "kms_key_pk": {
                "related_table": "kms_keys",
                "related_column": "key_arn",
                "this_table_column": "kms_key_id",
            },
        },
        "config_delivery_channels": {
            "s3_bucket_fk": {
                "related_table": "s3_buckets",
                "related_column": "name",
                "this_table_column": "s3_bucket_name",
            },
        },
        "cloudformation_stacks": {
            "parent_stack_fk": {
                "related_table": "cloudformation_stacks",
                "related_column": "stack_id",
                "this_table_column": "parent_id",
            },
            "iam_role_fk": {
                "related_table": "iam_roles",
                "related_column": "arn",
                "this_table_column": "role_arn",
            },
        },
        "neptune_db_instances": {
            "vpc_security_groups": {
                "many_to_many": True,
                "m2m_table_name": "neptune_instance_security_group",
                "related_table": "ec2_security_groups",
                "related_column": "group_id",
                "custom_join_query": """, json_array_elements(vpc_security_groups::json) AS e JOIN ec2_security_groups ec2_security_groups_tmp ON (e ->> 'VpcSecurityGroupId') = ec2_security_groups_tmp.group_id """,
            },
            "kms_key_pk": {
                "related_table": "kms_keys",
                "related_column": "key_arn",
                "this_table_column": "kms_key_id",
            },
        }
    },
)
# fmt: on



from typing import Any
from typing import Optional

from botocore import xform_name
from botocore.loaders import Loader
from botocore.model import ServiceModel

from .constants import METAMAP
from .constants import SUBRESOURCES_METAMAP

spark_type_map = {"structure": "struct", "list": "array"}

loader = Loader()


class SparkAWSSchemaSerializer:
    def __init__(
        self,
        custom_types: Optional[dict] = None,
    ):
        self.custom_types = custom_types or {}
        self.meta_map = (METAMAP | SUBRESOURCES_METAMAP)

    def schema(self, metadata_type: str = "ec2_instances"):

        service = metadata_type.split("_")[0]
        operation_name: Any = self.meta_map[metadata_type]["fetch_method"]

        # TODO avoid hardcoding service-2
        json_service_model = loader.load_service_model(service_name=service, type_name="service-2")
        sm = ServiceModel(json_service_model, service_name=service)

        mapping = {}
        for op_name in sm.operation_names:  # type: ignore
            py_operation_name = xform_name(op_name)
            mapping[py_operation_name] = op_name

        operation_model = sm.operation_model(mapping[operation_name])
        shape = operation_model.output_shape
        return self._get_shape_and_return_schema(shape=shape, metadata_type=metadata_type)

    def _get_shape_and_return_schema(
        self,
        shape,
        metadata_type,
    ):
        page_key = self.meta_map[metadata_type].get("page_key")
        response_key: Any = self.meta_map[metadata_type].get("response_key")

        if page_key:
            inner_response = shape.members[page_key]

            if inner_response.type_name == "structure":
                response_shape = inner_response.members[response_key].member
            elif inner_response.type_name == "list":
                response_shape = inner_response.member.members[response_key].member
            else:
                raise ValueError(shape)

            return self.parse(shape=response_shape)

        else:
            inner_response = shape

            if response_key:
                inner_response = inner_response.members[response_key]

            if inner_response.type_name == "list":
                response_shape = inner_response.member
            else:
                response_shape = inner_response

            if response_shape.type_name in ["string", "map"]:
                fields = [self.parse(inner_response, name=response_key)]
                return {
                    "type": "struct",
                    "fields": fields,
                }

            return self.parse(response_shape)

    def _parse_shape_blob(self, shape, name):
        return {
            "name": name,
            "type": "binary",
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_long(self, shape, name):
        return {
            "name": name,
            "type": self.custom_types.get("long", "long"),
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_double(self, shape, name):
        return {
            "name": name,
            "type": self.custom_types.get("double", "double"),
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_float(self, shape, name):
        return {
            "name": name,
            "type": "float",
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_map(self, shape, name):
        return {
            "name": name,
            "type": {
                "type": "map",
                "keyType": shape.key.type_name,
                "valueType": shape.value.type_name,
                "valueContainsNull": True,
            },
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_integer(self, shape, name):
        return {
            "name": name,
            "type": "integer",
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_boolean(self, shape, name):
        return {
            "name": name,
            "type": "boolean",
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_string(self, shape, name):
        output = {
            "name": name,
            "type": self.custom_types.get("string", "string"),
            "nullable": True,
            "metadata": {},
        }

        return output

    def _parse_shape_timestamp(self, shape, name):
        return {
            "name": name,
            "type": "timestamp",
            "nullable": True,
            "metadata": {},
        }

    def _parse_shape_list(self, shape, name):
        if shape.member.type_name not in ["structure", "list"]:
            element_type = shape.member.type_name
        else:
            element_type = self.parse(shape.member)

        output = {
            "type": {
                "type": "array",
                "elementType": element_type,
                "containsNull": True,
            },
            "nullable": True,
            "metadata": {},
            "name": name,
        }

        return output

    def _parse_shape_structure(self, shape, name):
        response = {"type": "struct"}

        if name:
            response["name"] = name

        fields = []

        for key, internal_shape in shape.members.items():
            if internal_shape.type_name not in ["structure", "list"]:
                item = self.parse(internal_shape, name=key)

            elif internal_shape.type_name == "structure":
                item = {"name": key, "type": self.parse(internal_shape), "nullable": True, "metadata": {}}

            elif internal_shape.type_name == "list":
                item = self.parse(shape=internal_shape, name=key)

            else:
                raise ValueError(internal_shape.type_name)

            fields.append(item)

        response["fields"] = fields  # type: ignore
        return response

    def parse(self, shape, name=None):
        type_name = shape.type_name
        return getattr(self, f"_parse_shape_{type_name}")(shape, name)



import boto3
import pytest

from nops_metadata import MetaFetcher


@pytest.fixture
def aws_session() -> boto3.Session:
    return boto3.Session()


@pytest.fixture()
def metafetcher(aws_session) -> MetaFetcher:
    return MetaFetcher(session=aws_session)



import json
from string import capwords
from typing import Any
from typing import Optional
from typing import Iterator

import boto3
from botocore.paginate import Paginator
from pyrsistent import thaw

from nops_metadata.constants import APPLY_NESTED_ELEMENT_PROCESSING


def _extract_keyed_page(page: dict, page_key: str, response_key: str) -> Iterator[dict[str, Any]]:
    keyed_page = page[page_key]

    if isinstance(keyed_page, dict):
        for element in keyed_page.get(response_key, []):
            yield element
    else:
        for section in keyed_page:
            for element in section[response_key]:
                yield element


def _handle_page(page: dict, page_key: str, response_key: str) -> Iterator[dict[str, Any]]:
    page_iter = (
        page[response_key]
        if not page_key
        else list(_extract_keyed_page(page, page_key=page_key, response_key=response_key))
    )

    if isinstance(page_iter, list):
        if page_iter and isinstance(page_iter[0], str):
            yield {response_key: page_iter}
        else:
            for element in page_iter:
                yield element
    else:
        if isinstance(page_iter, str):
            yield {response_key: page_iter}
        else:
            yield page_iter


def process_element_list(input_list):
    for item in input_list:
        if "AssumeRolePolicyDocument" in item:
            item = process_element(item)


def process_element(element, metadata_type: Optional[str] = None):
    for field in APPLY_NESTED_ELEMENT_PROCESSING.get(metadata_type, []):
        if isinstance(element[field], list):
            process_element_list(element[field])
        # TODO: Add process_element_dict once it needed

    if "AssumeRolePolicyDocument" in element:
        element["AssumeRolePolicyDocument"] = json.dumps(element["AssumeRolePolicyDocument"])
    return element


def resource_listing(
    session: boto3.Session,
    metaname: str,
    fetch_method: str,
    response_key: str,
    page_key: str,
    call_kwargs: dict[str, Any],
    region_name: str,
) -> Iterator[dict]:
    service = metaname.split("_")[0]
    client_kwargs = {}
    if region_name:
        client_kwargs["region_name"] = region_name

    client = session.client(service, **client_kwargs)

    if client.can_paginate(fetch_method):
        paginator: Paginator = client.get_paginator(fetch_method)

        for page in paginator.paginate(**thaw(call_kwargs)):
            for element in _handle_page(page=page, page_key=page_key, response_key=response_key):
                yield process_element(element, metadata_type=metaname)
    else:
        response: list[dict] = getattr(client, fetch_method)(**call_kwargs)

        if response_key:
            response = response[response_key]

        if isinstance(response, list):
            for el in response:
                yield el
        else:
            yield response



from datetime import datetime
from typing import Any
from typing import Optional

from botocore import xform_name
from botocore.loaders import Loader
from botocore.model import ServiceModel
from pydantic import BaseModel
from pydantic import create_model

from .constants import METAMAP

loader = Loader()


class PydanticSchemaGenerator:
    def __init__(
        self,
        custom_types: Optional[dict] = None,
    ):
        self.custom_types = custom_types or {}
        self.type_store = {}

    def schema(self, metadata_type: str = "ec2_instances"):
        self.type_store = {}  # Reset type store to avoid conflicts with other schemas.
        service = metadata_type.split("_")[0]
        operation_name: Any = METAMAP[metadata_type]["fetch_method"]

        json_service_model = loader.load_service_model(service_name=service, type_name="service-2")
        sm = ServiceModel(json_service_model, service_name=service)

        mapping = {}
        for op_name in sm.operation_names:  # type: ignore
            py_operation_name = xform_name(op_name)
            mapping[py_operation_name] = op_name

        operation_model = sm.operation_model(mapping[operation_name])
        shape = operation_model.output_shape
        return self._get_shape_and_return_schema(shape=shape, metadata_type=metadata_type)

    def _get_shape_and_return_schema(
        self,
        shape,
        metadata_type,
    ):
        page_key = METAMAP[metadata_type].get("page_key")
        response_key: Any = METAMAP[metadata_type]["response_key"]

        if page_key:
            inner_response = shape.members[page_key]

            if inner_response.type_name == "structure":
                response_shape = inner_response.members[response_key].member
            elif inner_response.type_name == "list":
                response_shape = inner_response.member.members[response_key].member
            else:
                raise ValueError(shape)

            return self.parse(shape=response_shape, name=page_key)

        else:
            inner_response = shape.members[response_key]

            if inner_response.type_name == "list":
                response_shape = inner_response.member
            else:
                response_shape = inner_response

            if response_shape.type_name in ["string"]:
                field = self.parse(inner_response, name=response_key)
                return create_model(inner_response.name, **{inner_response.name: field})

            response = self.parse(response_shape, name=response_key)
            if isinstance(response, tuple):
                return create_model(response_key, **{response_key: response})

            return response

    def _parse_shape_blob(self, shape, name, *args):
        return (bytes, None)

    def _parse_shape_long(self, shape, name, *args):
        return (int, None)

    def _parse_shape_double(self, shape, name, *args):
        return (float, None)

    def _parse_shape_float(self, shape, name, *args):
        return (float, None)

    def _parse_shape_map(self, shape, name, *args):
        return (dict, None)

    def _parse_shape_integer(self, shape, name, *args):
        return (int, None)

    def _parse_shape_boolean(self, shape, name, *args):
        return (bool, None)

    def _parse_shape_string(self, shape, name, *args):
        return (str, None)

    def _parse_shape_timestamp(self, shape, name, *args):
        return (datetime, None)

    def _parse_shape_list(self, shape, name, previous_shape_name):
        array_type = self.parse(shape=shape.member, name=name, previous_shape_name=previous_shape_name)
        response = (list[array_type], None)
        return response

    def _parse_shape_structure(self, shape, name, previous_shape_name=""):
        class_attrs = {}
        for key, internal_shape in shape.members.items():
            item = self.parse(shape=internal_shape, name=key, previous_shape_name=name)
            if not isinstance(item, tuple):
                item = (item, None)
            class_attrs[key] = item

        type_name = f"{previous_shape_name}{name}"

        if type_name not in self.type_store:
            self.type_store[type_name] = create_model(f"{previous_shape_name}{name}", **class_attrs)

        response = self.type_store[type_name]
        return response

    def parse(self, shape, name="", previous_shape_name=""):
        return getattr(self, f"_parse_shape_{shape.type_name}")(shape, name, previous_shape_name)



import logging
from itertools import chain
from pprint import pprint
from uuid import uuid4

from pyspark.sql import SparkSession
from pyspark.sql.types import StructType

from nops_metadata import MetaFetcher


def test_spark_schema_output(metafetcher: MetaFetcher):
    spark = SparkSession.builder.appName(f"testing_metafetcher_{uuid4()}").getOrCreate()

    for metadata_type in metafetcher.metadata_types:
        schema = metafetcher.schema(metadata_type=metadata_type)
        spark_schema = StructType.fromJson(schema)
        df = spark.createDataFrame(data=[], schema=spark_schema)
        assert df


def test_pydantic_schema_output(metafetcher: MetaFetcher):
    def _getter(metadata_type, region):
        try:
            for element in metafetcher.fetch(metadata_type=metadata_type, region_name=region):
                yield element
        except Exception as e:
            logging.warning(e)

    for metadata_type in metafetcher.metadata_types:

        # if metadata_type not in debug:
        #     continue

        try:
            regions = metafetcher.metadata_regions(metadata_type=metadata_type)
            regions.insert(0, "us-west-2")

            gens = [_getter(metadata_type, region) for region in regions]
            response = chain(*gens)
            single_resource = next(response)

            schema = metafetcher.schema(metadata_type=metadata_type, schema_type="pydantic")
            processed_item = schema(**single_resource)
            assert schema.schema()["properties"]
            assert processed_item

        except StopIteration:
            pass
